import erlab.analysis as era
import numpy as np
import xarray as xr


def test_gaussian_filter():
    # Create a test input DataArray
    darr = xr.DataArray(np.arange(50, step=2).reshape((5, 5)), dims=["x", "y"])

    # Define the expected output
    expected_output = xr.DataArray(
        np.array(
            [
                [3, 5, 7, 8, 10],
                [10, 12, 14, 15, 17],
                [20, 22, 24, 25, 27],
                [29, 31, 33, 34, 36],
                [36, 38, 40, 41, 43],
            ]
        ),
        dims=["x", "y"],
    )

    # Apply the gaussian_filter function
    result = era.image.gaussian_filter(darr, sigma={"x": 1.0, "y": 1.0})

    # Check if the result matches the expected output
    assert np.allclose(result, expected_output)


def test_gaussian_laplace():
    # Create a test input DataArray
    darr = xr.DataArray(np.arange(50, step=2).reshape((5, 5)), dims=["x", "y"])

    # Define the expected output
    expected_output = xr.DataArray(
        np.array(
            [
                [4, 4, 4, 4, 4],
                [2, 2, 2, 2, 2],
                [0, 0, 0, 0, 0],
                [-2, -2, -2, -2, -2],
                [-4, -4, -4, -4, -4],
            ]
        ),
        dims=["x", "y"],
    )

    # Apply the gaussian_laplace function
    result = era.image.gaussian_laplace(darr, sigma={"x": 1.0, "y": 1.0})

    # Check if the result matches the expected output
    assert np.allclose(result, expected_output)

    # Additional test case
    # Define the expected output
    expected_output2 = xr.DataArray(
        np.array(
            [
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, 0],
                [-1, -1, -1, -1, -1],
                [-1, -1, -1, -1, -1],
            ]
        ),
        dims=["x", "y"],
    )

    # Apply the gaussian_laplace function
    result2 = era.image.gaussian_laplace(darr, sigma=2.0)

    # Check if the result matches the expected output
    assert np.allclose(result2, expected_output2)


def test_laplace():
    # Create a test input DataArray
    darr = xr.DataArray(np.arange(50, step=2).reshape((5, 5)), dims=["x", "y"])

    # Define the expected output
    expected_output = xr.DataArray(
        np.array(
            [
                [12, 10, 10, 10, 8],
                [2, 0, 0, 0, -2],
                [2, 0, 0, 0, -2],
                [2, 0, 0, 0, -2],
                [-8, -10, -10, -10, -12],
            ]
        ),
        dims=["x", "y"],
    )

    # Apply the laplace function
    result = era.image.laplace(darr)

    # Check if the result matches the expected output
    assert np.allclose(result, expected_output)


def test_minimum_gradient():
    # Create a test input DataArray
    darr = xr.DataArray(np.arange(50, step=2).reshape((5, 5)), dims=["x", "y"])

    # Define the expected output
    expected_output = xr.DataArray(
        np.array(
            [
                [0.0, 0.00283506, 0.00567012, 0.00850517, 0.01215542],
                [0.01031404, 0.01225726, 0.01430013, 0.01634301, 0.01856527],
                [0.02062807, 0.02247164, 0.02451452, 0.02655739, 0.0288793],
                [0.03094211, 0.03268602, 0.0347289, 0.03677177, 0.03919334],
                [0.06077708, 0.05953621, 0.06237127, 0.06520633, 0.06622662],
            ]
        ),
        dims=["x", "y"],
    )

    # Apply the minimum_gradient function
    result = era.image.minimum_gradient(darr).astype(np.float32)

    # Check if the result matches the expected output
    assert np.allclose(result, expected_output)


def test_scaled_laplace():
    # Create a test input DataArray
    darr = xr.DataArray(
        np.arange(50, step=2).reshape((5, 5)).astype(float), dims=["x", "y"]
    )

    # Define the expected output
    expected_output = xr.DataArray(
        np.array(
            [
                [12.0, 10.0, 10.0, 10.0, 8.0],
                [2.0, 0.0, 0.0, 0.0, -2.0],
                [2.0, 0.0, 0.0, 0.0, -2.0],
                [2.0, 0.0, 0.0, 0.0, -2.0],
                [-8.0, -10.0, -10.0, -10.0, -12.0],
            ]
        ),
        dims=["x", "y"],
    )

    # Apply the scaled_laplace function
    result = era.image.scaled_laplace(darr)

    # Check if the result matches the expected output
    assert np.allclose(result, expected_output)


def test_curvature():
    # Create a test input DataArray
    darr = xr.DataArray(np.arange(50, step=2).reshape((5, 5)), dims=["x", "y"]) ** 2

    # Define the expected output
    expected_output = xr.DataArray(
        np.array(
            [
                [0.11852942, 0.11855069, 0.11778077, 0.11184719, 0.10558571],
                [0.16448492, 0.16107288, 0.15683689, 0.14772279, 0.1385244],
                [0.17403091, 0.16649966, 0.15876051, 0.14719689, 0.1361662],
                [0.09486956, 0.09038468, 0.08598027, 0.0783563, 0.07130584],
                [0.05139264, 0.04942051, 0.04746361, 0.04241876, 0.03781145],
            ]
        ),
        dims=["x", "y"],
    )

    # Apply the curvature function
    result = era.image.curvature(darr).astype(np.float32)

    # Check if the result matches the expected output
    assert np.allclose(result, expected_output)



"""Interactive tool for visualizing dispersive data."""

__all__ = ["dtool"]

import functools
import os
import sys

import numpy as np
import pyqtgraph as pg
import varname
import xarray as xr
from qtpy import QtCore, QtWidgets, uic

import erlab.analysis
from erlab.analysis.image import (
    curvature,
    gaussian_filter,
    minimum_gradient,
    scaled_laplace,
)
from erlab.interactive.utilities import (
    copy_to_clipboard,
    gen_function_code,
    parse_data,
    xImageItem,
)


class DerivativeTool(
    *uic.loadUiType(os.path.join(os.path.dirname(__file__), "dtool.ui"))
):
    def __init__(self, data: xr.DataArray, *, data_name: str | None = None):
        if data_name is None:
            try:
                data_name = varname.argname("data", func=self.__init__, vars_only=False)
            except varname.VarnameRetrievingError:
                data_name = "data"

        self.data_name: str = data_name

        # Initialize UI
        super().__init__()
        self.setupUi(self)
        self.setWindowTitle("")

        if data.ndim != 2:
            raise ValueError("Input DataArray must be 2D")

        self.data: xr.DataArray = parse_data(data)
        self._result: xr.DataArray = self.data.copy()

        self.xdim: str = self.data.dims[1]
        self.ydim: str = self.data.dims[0]

        self.xinc: float = abs(float(self.data[self.xdim][1] - self.data[self.xdim][0]))
        self.yinc: float = abs(float(self.data[self.ydim][1] - self.data[self.ydim][0]))

        self.sx_spin.setRange(0.1 * self.xinc, 50 * self.xinc)
        self.sy_spin.setRange(0.1 * self.yinc, 50 * self.yinc)

        self.sx_spin.setSingleStep(self.xinc / 2)
        self.sy_spin.setSingleStep(self.yinc / 2)

        self.reset_smooth()
        self.reset_interp()

        self.interp_group.setChecked(False)

        self.images: tuple[xImageItem, xImageItem] = (
            xImageItem(axisOrder="row-major"),
            xImageItem(axisOrder="row-major"),
        )
        self.hists: tuple[pg.HistogramLUTItem, pg.HistogramLUTItem] = (
            pg.HistogramLUTItem(),
            pg.HistogramLUTItem(),
        )
        self.plots: list[pg.PlotItem] = []

        for i in range(2):
            plot = self.graphics_layout.addPlot(i, 0, 1, 1)
            self.plots.append(plot)
            self.graphics_layout.addItem(self.hists[i], i, 1)
            plot.addItem(self.images[i])
            self.hists[i].setImageItem(self.images[i])
            # plot.showGrid(x=True, y=True, alpha=0.5)

        self.plots[0].setXLink(self.plots[1])
        self.plots[0].setYLink(self.plots[1])

        self.interp_rst_btn.clicked.connect(self.reset_interp)
        self.smooth_rst_btn.clicked.connect(self.reset_smooth)

        self.interp_group.toggled.connect(self.update_preprocess)
        self.smooth_group.toggled.connect(self.update_preprocess)
        self.nx_spin.valueChanged.connect(self.update_preprocess)
        self.ny_spin.valueChanged.connect(self.update_preprocess)
        self.sx_spin.valueChanged.connect(self.update_preprocess)
        self.sy_spin.valueChanged.connect(self.update_preprocess)
        self.sn_spin.valueChanged.connect(self.update_preprocess)

        self.hi_spin.valueChanged.connect(self.update_image)
        self.lo_spin.valueChanged.connect(self.update_image)

        self.tab_widget.currentChanged.connect(self.update_result)
        self.x_radio.clicked.connect(self.update_result)
        self.y_radio.clicked.connect(self.update_result)
        self.lapl_factor_spin.valueChanged.connect(self.update_result)
        self.curv_a0_spin.valueChanged.connect(self.update_result)
        self.curv_factor_spin.valueChanged.connect(self.update_result)

        self.copy_btn.clicked.connect(self.copy_code)

        self.update_result()

    @QtCore.Slot()
    def reset_smooth(self):
        self.sx_spin.setValue(self.xinc)
        self.sy_spin.setValue(self.yinc)

    @QtCore.Slot()
    def reset_interp(self):
        self.nx_spin.setValue(self.data.sizes[self.xdim])
        self.ny_spin.setValue(self.data.sizes[self.ydim])

    @property
    def result(self) -> xr.DataArray:
        return self._result

    @result.setter
    def result(self, value: xr.DataArray):
        self._result = value
        self.update_image()

    @functools.cached_property
    def processed_data(self) -> xr.DataArray:
        out = self.data
        if self.interp_group.isChecked():
            out = self.data.interp(
                {
                    self.xdim: np.linspace(
                        *self.data[self.xdim][[0, -1]], self.nx_spin.value()
                    ),
                    self.ydim: np.linspace(
                        *self.data[self.ydim][[0, -1]], self.ny_spin.value()
                    ),
                }
            )
        if self.smooth_group.isChecked():
            for _ in range(self.sn_spin.value()):
                out = gaussian_filter(
                    out,
                    sigma={
                        self.xdim: np.round(
                            self.sx_spin.value(), self.sx_spin.decimals()
                        ),
                        self.ydim: np.round(
                            self.sy_spin.value(), self.sy_spin.decimals()
                        ),
                    },
                )
        self.images[0].setDataArray(out)
        return out

    @QtCore.Slot()
    def update_image(self):
        self.images[1].setDataArray(
            self.result, levels=self.get_levels(self.result.values)
        )

    def get_levels(self, data, cutoff=None) -> tuple[float, float]:
        if cutoff is None:
            cutoff = (self.lo_spin.value(), self.hi_spin.value())
        else:
            try:
                cutoff = list(cutoff.__iter__)
            except AttributeError:
                cutoff = [cutoff] * 2

        pu, pl = np.percentile(data, [100 - cutoff[1], cutoff[0]])
        return max(pl, data.min()), min(pu, data.max())

    @QtCore.Slot()
    def update_preprocess(self):
        self.__dict__.pop("processed_data", None)
        self.update_result()

    @QtCore.Slot()
    def update_result(self):
        match self.tab_widget.currentIndex():
            case 0:
                dim = self.xdim if self.x_radio.isChecked() else self.ydim
                self.result = self.processed_data.differentiate(dim).differentiate(dim)
            case 1:
                self.result = scaled_laplace(
                    self.processed_data,
                    factor=np.round(
                        self.lapl_factor_spin.value(), self.lapl_factor_spin.decimals()
                    ),
                )
            case 2:
                self.result = curvature(
                    self.processed_data,
                    a0=np.round(
                        self.curv_a0_spin.value(), self.curv_a0_spin.decimals()
                    ),
                    factor=np.round(
                        self.curv_factor_spin.value(), self.curv_factor_spin.decimals()
                    ),
                )
            case 3:
                self.result = minimum_gradient(self.processed_data)

    def copy_code(self):
        lines: list[str] = []

        data_name = (
            self.data_name
        )  # "".join([s.strip() for s in self.data_name.split("\n")])
        if self.interp_group.isChecked():
            arg_dict = {
                dim: f"|np.linspace(*{data_name}['{dim}'][[0, -1]], {n})|"
                for dim, n in zip(
                    [self.xdim, self.ydim], [self.nx_spin.value(), self.ny_spin.value()]
                )
            }
            lines.append(
                gen_function_code(
                    copy=False, **{f"_processed = {data_name}.interp": [arg_dict]}
                )
            )
            data_name = "_processed"

        if self.smooth_group.isChecked():
            arg_dict = {
                "sigma": dict(
                    zip(
                        (self.xdim, self.ydim),
                        [
                            np.round(s.value(), s.decimals())
                            for s in (self.sx_spin, self.sy_spin)
                        ],
                    )
                )
            }
            lines.append(f"_processed = {data_name}.copy()")
            data_name = "_processed"
            lines.append(f"for _ in range({self.sn_spin.value()}):")
            lines.append(
                "\t"
                + gen_function_code(
                    copy=False,
                    **{
                        "_processed = era.image.gaussian_filter": [
                            f"|{data_name}|",
                            arg_dict,
                        ]
                    },
                )
            )
            data_name = "_processed"

        if self.tab_widget.currentIndex() == 0:
            dim = self.xdim if self.x_radio.isChecked() else self.ydim
            lines.append(f"{data_name}.differentiate('{dim}').differentiate('{dim}')")
        else:
            match self.tab_widget.currentIndex():
                case 1:
                    fname = "era.image.scaled_laplace"
                    arg_dict = {
                        "factor": np.round(
                            self.lapl_factor_spin.value(),
                            self.lapl_factor_spin.decimals(),
                        )
                    }
                case 2:
                    fname = "era.image.curvature"
                    arg_dict = {
                        "a0": np.round(
                            self.curv_a0_spin.value(), self.curv_a0_spin.decimals()
                        ),
                        "factor": np.round(
                            self.curv_factor_spin.value(),
                            self.curv_factor_spin.decimals(),
                        ),
                    }
                case 3:
                    fname = "era.image.minimum_gradient"
                    arg_dict = {}

            lines.append(
                gen_function_code(
                    copy=False, **{f"result = {fname}": [f"|{data_name}|", arg_dict]}
                )
            )

        copy_to_clipboard(lines)


def dtool(data, data_name: str | None = None, *, execute: bool | None = None):
    if data_name is None:
        try:
            data_name = varname.argname("data", func=dtool, vars_only=False)
        except varname.VarnameRetrievingError:
            data_name = "data"

    qapp = QtWidgets.QApplication.instance()
    if not qapp:
        qapp = QtWidgets.QApplication(sys.argv)

    qapp.setStyle("Fusion")

    win = DerivativeTool(data, data_name=data_name)
    win.show()
    win.raise_()
    win.activateWindow()

    if execute is None:
        execute = True
        try:
            shell = get_ipython().__class__.__name__  # type: ignore
            if shell in ["ZMQInteractiveShell", "TerminalInteractiveShell"]:
                execute = False
                from IPython.lib.guisupport import start_event_loop_qt4

                start_event_loop_qt4(qapp)
        except NameError:
            pass
    if execute:
        qapp.exec()


if __name__ == "__main__":
    import erlab.io

    dat = erlab.io.load_wave("/Users/khan/Downloads/dose5_k.ibw").T.sel(
        W=slice(-0.7, 0.7)
    )
    dtool(dat)



import numpy as np
import pytest
import scipy.special
from erlab.analysis.fit.functions.general import (
    _gen_kernel,
    do_convolve,
    fermi_dirac,
    fermi_dirac_linbkg,
    fermi_dirac_linbkg_broad,
    gaussian_wh,
    lorentzian_wh,
    step_broad,
    step_linbkg_broad,
)

KB_EV = 8.617333262145179e-5


def test_gen_kernel():
    # Define test input values
    x = np.array([0, 1, 2, 3, 4])
    resolution = 2.0
    pad = 5

    # Call the function to generate the kernel
    extended, gauss = _gen_kernel(x, resolution, pad)

    # Define the expected results
    expected_extended = np.arange(-4, 9, 1, dtype=np.float64)
    expected_gauss = np.array(
        [
            7.16733764e-06,
            9.17419217e-04,
            2.93574150e-02,
            2.34859320e-01,
            4.69718639e-01,
            2.34859320e-01,
            2.93574150e-02,
            9.17419217e-04,
            7.16733764e-06,
        ]
    )

    # Check if the generated extended and gauss arrays match the expected results
    assert np.allclose(extended, expected_extended)
    assert np.allclose(gauss, expected_gauss)


def test_do_convolve():
    # Test case 1: Convolve a linear function with a Gaussian kernel
    x = np.linspace(0, 10, 5)

    def testfunc(x):
        return x

    resolution = 1.0
    expected = np.array(
        [-6.61744490e-24, 5.87148334e00, 1.17429667e01, 1.76144500e01, 2.34859334e01]
    )
    result = do_convolve(x, testfunc, resolution)
    assert np.allclose(result, expected)

    # Test case 2: Convolve a sine function with a Gaussian kernel
    x = np.linspace(0, 2 * np.pi, 5)

    def testfunc(x):
        return np.sin(x)

    resolution = 0.5
    expected = np.array(
        [0.00000000e00, 2.95132925e00, 3.61433592e-16, -2.95132925e00, -7.22867185e-16]
    )
    result = do_convolve(x, testfunc, resolution)
    assert np.allclose(result, expected)


def test_gaussian_wh():
    x = np.linspace(0, 10, 100)
    center = 5.0
    width = 2.0
    height = 1.0
    expected_result = height * np.exp(
        -16 * np.log(2) * (1.0 * x - center) ** 2 / max(1e-10, width**2)
    )
    assert np.allclose(gaussian_wh(x, center, width, height), expected_result)


def test_lorentzian_wh():
    x = np.linspace(0, 10, 100)
    center = 5.0
    width = 2.0
    height = 1.0
    expected_result = height / (1 + 4 * ((1.0 * x - center) / max(1.0e-15, width)) ** 2)
    assert np.allclose(lorentzian_wh(x, center, width, height), expected_result)


@pytest.mark.filterwarnings("ignore:overflow encountered in exp.*:RuntimeWarning")
def test_fermi_dirac():
    x = np.linspace(0, 10, 100)
    center = 5.0
    temp = 2.0
    expected_result = 1 / (1 + np.exp((1.0 * x - center) / max(1.0e-15, temp * KB_EV)))
    assert np.allclose(fermi_dirac(x, center, temp), expected_result)


@pytest.mark.filterwarnings("ignore:overflow encountered in exp.*:RuntimeWarning")
def test_fermi_dirac_linbkg():
    x = np.linspace(0, 10, 100)
    center = 5.0
    temp = 2.0
    back0 = 1.0
    back1 = 0.5
    dos0 = 0.8
    dos1 = 0.3
    expected_result = (back0 + back1 * x) + (dos0 - back0 + (dos1 - back1) * x) / (
        1 + np.exp((1.0 * x - center) / max(1.0e-15, temp * KB_EV))
    )
    assert np.allclose(
        fermi_dirac_linbkg(x, center, temp, back0, back1, dos0, dos1), expected_result
    )


def test_fermi_dirac_linbkg_broad():
    x = np.linspace(0, 10, 100)
    center = 5.0
    temp = 300.0
    resolution = 0.1
    back0 = 0.1
    back1 = 0.2
    dos0 = 0.3
    dos1 = 0.4
    expected_result = do_convolve(
        x,
        fermi_dirac_linbkg,
        resolution=resolution,
        center=center,
        temp=temp,
        back0=back0,
        back1=back1,
        dos0=dos0,
        dos1=dos1,
    )
    assert np.allclose(
        fermi_dirac_linbkg_broad(x, center, temp, resolution, back0, back1, dos0, dos1),
        expected_result,
    )


def test_step_broad():
    x = np.linspace(-10, 10, 100)
    center = 0.0
    sigma = 1.0
    amplitude = 1.0
    expected_result = (
        amplitude
        * 0.5
        * scipy.special.erfc((1.0 * x - center) / max(1.0e-15, np.sqrt(2) * sigma))
    )
    assert np.allclose(step_broad(x, center, sigma, amplitude), expected_result)


def test_step_linbkg_broad():
    x = np.linspace(0, 10, 100)
    center = 5.0
    sigma = 2.0
    back0 = 1.0
    back1 = 0.5
    dos0 = 2.0
    dos1 = 1.0
    expected_result = (back0 + back1 * x) + (dos0 - back0 + (dos1 - back1) * x) * (
        step_broad(x, center, sigma, 1.0)
    )
    assert np.allclose(
        step_linbkg_broad(x, center, sigma, back0, back1, dos0, dos1), expected_result
    )



import numpy as np
import pytest
import xarray as xr
from erlab.analysis.fit.functions.dynamic import (
    FermiEdge2dFunction,
    MultiPeakFunction,
    PolynomialFunction,
)
from erlab.analysis.fit.functions.general import gaussian_wh, lorentzian_wh

RAND_STATE = np.random.RandomState(1)


def test_poly_func_call():
    # Test case 1: Evaluate polynomial function with numpy array input
    x = np.arange(5, dtype=np.float64)
    coeffs = RAND_STATE.randn(3)
    expected_result = np.polyval(np.asarray(list(reversed(coeffs))), x)
    assert np.allclose(
        PolynomialFunction(degree=2)(x, *coeffs),
        expected_result,
    )

    # Test case 2: Evaluate polynomial function with xarray input
    x = xr.DataArray(np.arange(5, dtype=np.float64), dims="x")
    coeffs = RAND_STATE.randn(3)
    expected_result = np.polyval(np.asarray(list(reversed(coeffs))), x)
    result = PolynomialFunction(degree=2)(x, *coeffs)
    assert isinstance(result, xr.DataArray)
    assert np.allclose(result, expected_result)

    # Test case 3: Evaluate polynomial function with parameters
    x = np.arange(5, dtype=np.float64)
    coeffs = RAND_STATE.randn(3)
    expected_result = np.polyval(np.asarray(list(reversed(coeffs))), x)
    params = dict(zip([f"c{i}" for i in range(3)], coeffs))
    result = PolynomialFunction(degree=2)(x, **params)
    assert np.allclose(result, expected_result)


def test_multi_peak_function_call():
    # Test case 1: Evaluate multi-peak function with numpy array input
    x = np.linspace(-5, 5, 20, dtype=np.float64)
    npeaks = 2
    peak_shapes = ["lorentzian", "gaussian"]
    fd = False
    convolve = False
    params = {
        "p0_center": 0.0,
        "p0_height": 1.0,
        "p0_width": 0.5,
        "p1_center": 2.0,
        "p1_height": 0.5,
        "p1_width": 0.2,
        "lin_bkg": 0.1,
        "const_bkg": 0.2,
    }
    expected_result = (
        lorentzian_wh(x, params["p0_center"], params["p0_width"], params["p0_height"])
        + gaussian_wh(x, params["p1_center"], params["p1_width"], params["p1_height"])
        + params["lin_bkg"] * x
        + params["const_bkg"]
    )
    assert np.allclose(
        MultiPeakFunction(npeaks, peak_shapes, fd, convolve)(x, **params),
        expected_result,
    )
    # Test case 2: With different peak shape signature
    assert np.allclose(
        MultiPeakFunction(npeaks, "l g", fd, convolve)(x, **params),
        expected_result,
    )

    # Test case 3: Evaluate multi-peak function with xarray input
    x = xr.DataArray(np.linspace(-5, 5, 20, dtype=np.float64), dims="x")
    result_xr = MultiPeakFunction(npeaks, peak_shapes, fd, convolve)(x, **params)
    assert isinstance(result_xr, xr.DataArray)
    assert np.allclose(result_xr, expected_result)

    # Test case 4: Evaluate multi-peak function with different peak shape signatures
    x = np.linspace(-5, 5, 20, dtype=np.float64)
    npeaks = 3
    peak_shapes = ["gaussian", "lorentzian", "gaussian"]
    fd = True
    convolve = True
    params = {
        "p0_center": 1.0,
        "p0_height": 0.5,
        "p0_width": 0.2,
        "p1_center": 2.0,
        "p1_height": 1.0,
        "p1_width": 0.5,
        "p2_center": 3.0,
        "p2_height": 0.8,
        "p2_width": 0.3,
        "lin_bkg": 0.2,
        "const_bkg": 0.1,
        "efermi": 0.0,
        "temp": 30,
        "offset": 0.5,
        "resolution": 0.1,
    }

    expected_result = MultiPeakFunction(npeaks, "lorentzian", fd, convolve)(x, **params)

    for peak_shapes in ("l l l", ["lorentzian"] * 3):
        assert np.allclose(
            MultiPeakFunction(npeaks, peak_shapes, fd, convolve)(x, **params),
            expected_result,
        )


@pytest.mark.filterwarnings("ignore:overflow encountered in exp.*:RuntimeWarning")
def test_fermi_edge_2d_function_call():
    # Difficult to test function values, just check if equal for numpy and xarray
    eV = np.linspace(-5, 5, 25, dtype=np.float64)
    alpha = RAND_STATE.randn(20)
    alpha.sort()
    params = {
        "c0": 1.0,
        "c1": 0.5,
        "lin_bkg": 0.1,
        "const_bkg": 0.2,
        "offset": 0.0,
        "temp": 30.0,
        "resolution": 0.02,
    }
    out_np = FermiEdge2dFunction()(eV, alpha, **params)

    eV = xr.DataArray(eV, dims="eV")
    alpha = xr.DataArray(alpha, dims="alpha")
    out_xr = FermiEdge2dFunction()(eV, alpha, **params)

    assert isinstance(out_xr, xr.DataArray)
    assert np.allclose(out_np.reshape((25, 20)), out_xr)



"""
Various image processing functions including tools for visualizing dispersive features.

Note
----
For scipy-based filter functions, the default value of the `mode` argument is 'nearest',
unlike the scipy default of 'reflect'.
"""

from collections.abc import Sequence

import numpy as np
import numpy.typing as npt
import scipy
import scipy.ndimage
import xarray as xr
from numba import carray, cfunc, types


def gaussian_filter(
    darr: xr.DataArray,
    sigma: float | dict[str, float] | Sequence[float],
    order: int | Sequence[int] | dict[str, int] = 0,
    mode: str | Sequence[str] | dict[str, str] = "nearest",
    cval: float = 0.0,
    truncate: float = 4.0,
    *,
    radius: None | float | Sequence[float] | dict[str, float] = None,
) -> xr.DataArray:
    """Coordinate-aware wrapper around `scipy.ndimage.gaussian_filter`.

    Parameters
    ----------
    darr
        The input DataArray.
    sigma
        The standard deviation(s) of the Gaussian filter in data dimensions. If a float,
        the same value is used for all dimensions, each scaled by the data step. If a
        dict, the value can be specified for each dimension using dimension names as
        keys. The filter is only applied to the dimensions specified in the dict. If a
        sequence, the values are used in the same order as the dimensions of the
        DataArray.
    order
        The order of the filter along each dimension. If an int, the same order is used
        for all dimensions. See Notes below for other options. Defaults to 0.
    mode
        The boundary mode used for the filter. If a str, the same mode is used for all
        dimensions. See Notes below for other options. Defaults to 'nearest'.
    cval
        Value to fill past edges of input if mode is 'constant'. Defaults to 0.0.
    truncate
        The truncation value used for the Gaussian filter. Defaults to 4.0.
    radius
        The radius of the Gaussian filter in data units. See Notes below. Defaults to
        None.

    Returns
    -------
    gaussian_filter : xarray.DataArray
        The filtered array with the same shape as the input DataArray.

    Note
    ----
    - The `sigma` and `radius` values should be in data coordinates, not pixels.
    - The input array is assumed to be regularly spaced.
    - `order`, `mode`, and `radius` can be specified for each dimension using a dict or
      a sequence. If a dict, the value can be specified for each dimension using
      dimension names as keys. If a sequence and `sigma` is given as a dictionary, the
      order is assumed to be the same as the keys in `sigma`. If `sigma` is not a
      dictionary, the order is assumed to be the same as the dimensions of the
      DataArray.

    See Also
    --------
    :func:`scipy.ndimage.gaussian_filter` : The underlying function used to apply the
        filter.

    Example
    -------
    >>> import numpy as np
    >>> import xarray as xr
    >>> import erlab.analysis as era
    >>> darr = xr.DataArray(np.arange(50, step=2).reshape((5, 5)), dims=["x", "y"])
    >>> darr
    <xarray.DataArray (x: 5, y: 5)> Size: 200B
    array([[ 0,  2,  4,  6,  8],
        [10, 12, 14, 16, 18],
        [20, 22, 24, 26, 28],
        [30, 32, 34, 36, 38],
        [40, 42, 44, 46, 48]])
    Dimensions without coordinates: x, y
    >>> era.image.gaussian_filter(darr, sigma=dict(x=1.0, y=1.0))
    <xarray.DataArray (x: 5, y: 5)> Size: 200B
    array([[ 3,  5,  7,  8, 10],
        [10, 12, 14, 15, 17],
        [20, 22, 24, 25, 27],
        [29, 31, 33, 34, 36],
        [36, 38, 40, 41, 43]])
    Dimensions without coordinates: x, y

    """
    if np.isscalar(sigma):
        sigma = {d: sigma for d in darr.dims}
    elif not isinstance(sigma, dict):
        sigma = dict(zip(darr.dims, sigma))

    # Get the axis indices to apply the filter
    axes = tuple(darr.get_axis_num(d) for d in sigma.keys())

    # Convert arguments to tuples acceptable by scipy
    if isinstance(order, dict):
        order = tuple(order.get(d, 0) for d in sigma.keys())
    if isinstance(mode, dict):
        mode = tuple(mode[d] for d in sigma.keys())
    if radius is not None:
        if len(radius) != len(sigma):
            raise ValueError("`radius` does not match dimensions of `sigma`")

        if np.isscalar(radius):
            radius = {d: radius for d in sigma.keys()}
        elif not isinstance(radius, dict):
            radius = dict(zip(sigma.keys(), radius))

        # Calculate radius in pixels
        radius: tuple[int, ...] = tuple(
            round(r / (darr[d].values[1] - darr[d].values[0]))
            for d, r in radius.items()
        )

    # Calculate sigma in pixels
    sigma: tuple[float, ...] = tuple(
        val / (darr[d].values[1] - darr[d].values[0]) for d, val in sigma.items()
    )

    return darr.copy(
        data=scipy.ndimage.gaussian_filter(
            darr.values,
            sigma=sigma,
            order=order,
            mode=mode,
            cval=cval,
            truncate=truncate,
            radius=radius,
            axes=axes,
        )
    )


def gaussian_laplace(
    darr: xr.DataArray,
    sigma: float | dict[str, float] | Sequence[float],
    mode: str | Sequence[str] | dict[str, str] = "nearest",
    cval: float = 0.0,
    **kwargs,
) -> xr.DataArray:
    """Coordinate-aware wrapper around `scipy.ndimage.gaussian_laplace`.

    This function calculates the Laplacian of the given array using Gaussian second
    derivatives.

    Parameters
    ----------
    darr
        The input DataArray.
    sigma
        The standard deviation(s) of the Gaussian filter in data dimensions. If a float,
        the same value is used for all dimensions, each scaled by the data step. If a
        dict, the value can be specified for each dimension using dimension names as
        keys. If a sequence, the values are used in the same order as the dimensions of
        the DataArray.
    mode
        The mode parameter determines how the input array is extended beyond its
        boundaries. If a string, the same mode is used for all dimensions. If a
        sequence, the values should be the modes for each dimension in the same order as
        the dimensions in the DataArray. If a dictionary, the keys should be dimension
        names and the values should be the corresponding modes, and every dimension in
        the DataArray must be present. Default is "nearest".
    cval
        Value to fill past edges of input if mode is 'constant'. Defaults to 0.0.
    **kwargs
        Additional keyword arguments to `scipy.ndimage.gaussian_filter`.

    Returns
    -------
    gaussian_laplace : xarray.DataArray
        The filtered array with the same shape as the input DataArray.

    Note
    ----
    - `sigma` should be in data coordinates, not pixels.
    - The input array is assumed to be regularly spaced.

    See Also
    --------
    :func:`scipy.ndimage.gaussian_laplace` : The underlying function used to apply the
        filter.
    """
    if np.isscalar(sigma):
        sigma = {d: sigma for d in darr.dims}
    elif not isinstance(sigma, dict):
        sigma = dict(zip(darr.dims, sigma))

    if len(sigma) != darr.ndim:
        raise ValueError(
            "`sigma` must be provided for every dimension of the DataArray"
        )

    # Convert mode to tuple acceptable by scipy
    if isinstance(mode, dict):
        mode = tuple(mode[d] for d in sigma.keys())

    # Calculate sigma in pixels
    sigma: tuple[float, ...] = tuple(
        val / (darr[d].values[1] - darr[d].values[0]) for d, val in sigma.items()
    )

    return darr.copy(
        data=scipy.ndimage.gaussian_laplace(
            darr.values, sigma=sigma, mode=mode, cval=cval, **kwargs
        )
    )


def gradient_magnitude(
    input: npt.NDArray[np.float64],
    dx: np.float64,
    dy: np.float64,
    mode: str = "nearest",
    cval: float = 0.0,
) -> npt.NDArray[np.float64]:
    r"""Calculate the gradient magnitude of an input array.

    The gradient magnitude is calculated as defined in Ref. :cite:p:`he2017mingrad`,
    using given :math:`\Delta x` and :math:`\Delta y` values.

    Parameters
    ----------
    input
        Input array.
    dx
        Step size in the x-direction.
    dy
        Step size in the y-direction.
    mode
        The mode parameter controls how the gradient is calculated at the boundaries.
        Default is 'nearest'. See `scipy.ndimage.generic_filter` for more information.
    cval
        The value to use for points outside the boundaries when mode is 'constant'.
        Default is 0.0. See `scipy.ndimage.generic_filter` for more information.

    Returns
    -------
    gradient_magnitude : numpy.ndarray
        Gradient magnitude of the input array. Has the same shape as :code:`input`.

    Note
    ----
    This function calculates the gradient magnitude of an input array by applying a
    filter to the input array using the given dx and dy values. The filter is defined by
    a kernel function that computes the squared difference between each element of the
    input array and the central element, divided by the corresponding distance value.
    The gradient magnitude is then calculated as the square root of the sum of the
    squared differences.
    """

    dxy = np.sqrt(dx**2 + dy**2)
    dist = np.array([[dxy, dy, dxy], [dx, 0.0, dx], [dxy, dy, dxy]]).flatten()

    @cfunc(
        types.intc(
            types.CPointer(types.float64),
            types.intp,
            types.CPointer(types.float64),
            types.voidptr,
        )
    )
    def _kernel(values_ptr, len_values, result, data):
        values = carray(values_ptr, (len_values,), dtype=types.float64)
        val = 0.0
        for i in range(9):
            if i != 4:
                val += ((values[i] - values[4]) / dist[i]) ** 2
        result[0] = np.sqrt(val)
        return 1

    # https://github.com/jni/llc-tools/issues/3#issuecomment-757134814
    func = scipy.LowLevelCallable(
        _kernel.ctypes, signature="int (double *, npy_intp, double *, void *)"
    )

    return scipy.ndimage.generic_filter(input, func, size=(3, 3), mode=mode, cval=cval)


def laplace(
    darr, mode: str | Sequence[str] | dict[str, str] = "nearest", cval: float = 0.0
) -> xr.DataArray:
    """Coordinate-aware wrapper around `scipy.ndimage.laplace`.

    This function calculates the Laplacian of the given array using approximate second
    derivatives.

    Parameters
    ----------
    darr
        The input DataArray.
    mode
        The mode parameter determines how the input array is extended beyond its
        boundaries. If a dictionary, the keys should be dimension names and the values
        should be the corresponding modes, and every dimension in the DataArray must be
        present. Otherwise, it retains the same behavior as in `scipy.ndimage.laplace`.
        Default is 'nearest'.
    cval
        Value to fill past edges of input if mode is 'constant'. Defaults to 0.0.

    Returns
    -------
    laplace : xarray.DataArray
        The filtered array with the same shape as the input DataArray.

    See Also
    --------
    :func:`scipy.ndimage.laplace` : The underlying function used to apply the filter.
    """
    if isinstance(mode, dict):
        mode = tuple(mode[d] for d in darr.dims)
    return darr.copy(data=scipy.ndimage.laplace(darr.values, mode=mode, cval=cval))


def minimum_gradient(
    darr: xr.DataArray, mode: str = "nearest", cval: float = 0.0
) -> xr.DataArray:
    """Minimum gradient method for detecting dispersive features in 2D data.

    The minimum gradient is calculated by dividing the input DataArray by the gradient
    magnitude. See Ref. :cite:p:`he2017mingrad`.

    Parameters
    ----------
    darr
        The 2D DataArray for which to calculate the minimum gradient.
    mode
        The mode parameter controls how the gradient is calculated at the boundaries.
        Default is 'nearest'. See `scipy.ndimage.generic_filter` for more information.
    cval
        The value to use for points outside the boundaries when mode is 'constant'.
        Default is 0.0. See `scipy.ndimage.generic_filter` for more information.

    Returns
    -------
    minimum_gradient : xarray.DataArray
        The minimum gradient of the input DataArray. Has the same shape as
        :code:`input`.

    Raises
    ------
    ValueError
        If the input DataArray is not 2D.

    Note
    ----
    - The input array is assumed to be regularly spaced.
    - Any zero gradient values are replaced with NaN.
    """

    if darr.ndim != 2:
        raise ValueError("DataArray must be 2D")

    xvals = darr[darr.dims[1]].values
    yvals = darr[darr.dims[0]].values

    dx = abs(xvals[1] - xvals[0])
    dy = abs(yvals[1] - yvals[0])

    grad = gradient_magnitude(
        darr.values.astype(np.float64), dx, dy, mode=mode, cval=cval
    )
    grad[np.isclose(grad, 0.0)] = np.nan
    return darr / darr.max(skipna=True) / grad


def scaled_laplace(
    darr,
    factor: float = 1.0,
    mode: str | Sequence[str] | dict[str, str] = "nearest",
    cval: float = 0.0,
) -> xr.DataArray:
    r"""Calculate the Laplacian of a 2D DataArray with different scaling for each axis.

    This function calculates the Laplacian of the given array using approximate second
    derivatives, taking the different scaling for each axis into account.

    .. math::
        \Delta f \sim \frac{\partial^2 f}{\partial x^2} \left(\frac{\Delta x}{\Delta y}\right)^{\!2} + \frac{\partial^2 f}{\partial y^2}

    See Ref. :cite:p:`zhang2011curvature` for more information.

    Parameters
    ----------
    darr
        The 2D DataArray for which to calculate the scaled Laplacian.
    factor
        The factor by which to scale the x-axis derivative. Negative values will scale
        the y-axis derivative instead. Default is 1.0.
    mode
        The mode parameter determines how the input array is extended beyond its
        boundaries. If a dictionary, the keys should be dimension names and the values
        should be the corresponding modes, and every dimension in the DataArray must be
        present. Otherwise, it retains the same behavior as in
        `scipy.ndimage.generic_laplace`. Default is 'nearest'.
    cval
        Value to fill past edges of input if mode is 'constant'. Defaults to 0.0.

    Returns
    -------
    scaled_laplace : xarray.DataArray
        The filtered array with the same shape as the input DataArray.

    Raises
    ------
    ValueError
        If the input DataArray is not 2D.

    Note
    ----
    The input array is assumed to be regularly spaced.

    See Also
    --------
    :func:`scipy.ndimage.generic_laplace` : The underlying function used to apply the
        filter.
    """
    if darr.ndim != 2:
        raise ValueError("DataArray must be 2D")

    xvals = darr[darr.dims[1]].values
    yvals = darr[darr.dims[0]].values

    dx = xvals[1] - xvals[0]
    dy = yvals[1] - yvals[0]
    weight = (dx / dy) ** 2

    if factor > 0:
        weight *= factor
    elif factor < 0:
        weight /= abs(factor)

    if isinstance(mode, dict):
        mode = tuple(mode[d] for d in darr.dims)

    def d2_scaled(input, axis, output, mode, cval):
        out = scipy.ndimage.correlate1d(input, [1, -2, 1], axis, output, mode, cval, 0)
        if axis == 1:
            out *= weight
        return out

    return darr.copy(
        data=scipy.ndimage.generic_laplace(darr.values, d2_scaled, mode=mode, cval=cval)
    )


def curvature(darr: xr.DataArray, a0: float = 1.0, factor: float = 1.0) -> xr.DataArray:
    """2D curvature method for detecting dispersive features.

    The curvature is calculated as defined by :cite:t:`zhang2011curvature`.

    Parameters
    ----------
    darr
        The 2D DataArray for which to calculate the curvature.
    a0
        The regularization constant. Reasonable values range from 0.001 to 10. Default
        is 1.0.
    factor
        The factor by which to scale the x-axis curvature. Negative values will scale
        the y-axis curvature instead. Default is 1.0.

    Returns
    -------
    curvature : xarray.DataArray
        The 2D curvature of the input DataArray. Has the same shape as :code:`input`.

    Raises
    ------
    ValueError
        If the input DataArray is not 2D.

    Note
    ----
    The input array is assumed to be regularly spaced.
    """

    if darr.ndim != 2:
        raise ValueError("DataArray must be 2D")

    xvals = darr[darr.dims[1]].values
    yvals = darr[darr.dims[0]].values

    dx = xvals[1] - xvals[0]
    dy = yvals[1] - yvals[0]
    weight = (dx / dy) ** 2

    if factor > 0:
        weight *= factor
    elif factor < 0:
        weight /= abs(factor)

    dfdx, dfdy = np.gradient(darr.values, axis=(1, 0))
    d2fdx2, d2fdydx = np.gradient(dfdx, axis=(1, 0))
    d2fdy2 = np.gradient(dfdy, axis=0)

    max_abs_dfdx_sq: float = np.max(np.abs(dfdx)) ** 2
    max_abs_dfdy_sq: float = np.max(np.abs(dfdy)) ** 2

    c0 = a0 * max(max_abs_dfdy_sq, weight * max_abs_dfdx_sq)

    curv = (
        (c0 + weight * dfdx**2) * d2fdy2
        - 2 * weight * dfdx * dfdy * d2fdydx
        + weight * (c0 + dfdy**2) * d2fdx2
    ) / (c0 + weight * dfdx**2 + dfdy**2) ** 1.5

    return darr.copy(data=curv)



import numpy as np
import xarray as xr
from erlab.analysis.utilities import shift


def test_shift():
    # Create a test input DataArray
    darr = xr.DataArray(
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(float), dims=["x", "y"]
    )

    # Create a test shift DataArray
    shift_arr = xr.DataArray([1, 0, 2], dims=["x"])

    # Perform the shift operation
    shifted = shift(darr, shift_arr, along="y")

    # Define the expected result
    expected = xr.DataArray(
        np.array([[np.nan, 1.0, 2.0], [4.0, 5.0, 6.0], [np.nan, np.nan, 7.0]]),
        dims=["x", "y"],
    )

    # Check if the shifted array matches the expected result
    assert np.allclose(shifted, expected, equal_nan=True)
